"""
    Module to run scripts from Telegram by calling custom commands
    For instant, good_morning will call the command defined in
    /script good_morning will call the command defined in [SCRIPTS] good_morning
"""
import logging
import pathlib
import shlex
import subprocess as sp

from pybliotecario.components.component_core import Component

logger = logging.getLogger(__name__)


def _parse_cmd_args(text, shell=False):
    """Separate script name and its arguments
    If shell == True, then we want to preserve quotations
    """
    split_text = shlex.split(text, posix=not shell)
    script = split_text[0]
    args = split_text[1:]
    return script, args


class Script(Component):
    """
    Run scripts in the remote system using `subprocess.run`

    This is a security-critical feature of the pybliotecario, so you should
    be very careful on how you use it. By default, only the `chat_id` accepted
    chat is able to run scripts in order to avoid any person to run scripts in your server.

    In order to add scripts that you want to run to the pybliotecario you have
    to add a [SCRIPT] section to the config file (by default pybliotecario.ini)
    with
        script_alias=/path/to/the/script
    then you can call the script from your telegram by calling:
        /script script_alias <any possible arguments>

    Two security (or insecurity rather) features are included in this component:
        - shell
        - everyone
    These are two options that can be added to the [SCRIPT] section of the config file
    and are off by default:
        - shell: with shell=True, the scripts are called through a system shell
        this is sometimes useful (if you want access to any shell-feature) but it could
        potentially allow any attackers to run _any_ commands in your server.
        On the other hand it also allows _you_ to run any commands.
        - everyone: with everyone=True, the `chat_id` is not checked

    In order to avoid potentially destructive behaviour shell and everyone can not be set
    simultaneously. If you want to do so you will have to change the code manually.
    """

    key_name = "SCRIPT"
    help_text = """ > Script module
    /script list: list all possible scripts
    /script script_name: execute the given script """

    def __init__(self, telegram_object, **kwargs):
        super().__init__(telegram_object, **kwargs)
        self.blocked = False
        script_section = self.read_config_section()
        self.scripts = dict(script_section)

        if not self.scripts:
            self.blocked = True
            return

        # Read the config section for the given keys
        # and if they do exist, remove them from self.scripts
        # I know it is confusing but the previous version was apparently python >3.7 only :(
        self._run_in_shell = self._bool_and_pop(script_section, "shell")
        self._allow_everyone = self._bool_and_pop(script_section, "everyone")

        if self._allow_everyone and self._run_in_shell:
            # Comment this for potentially destructive behaviour
            logger.error("The options `everyone` and `shell` can not be used at the same time!")
            self.send_msg("The script module is blocked due to the use of everyone and shell")
            self.blocked = True

        default_keys = set(self.configuration["DEFAULT"].keys())
        self.script_names = list(set(self.scripts.keys()) - default_keys)

    def _bool_and_pop(self, section, key):
        """Check whether a key exists, read it as a boolean
        and pop-it-out"""
        if key in section:
            try:
                val = section.getboolean(key)
                self.scripts.pop(key)
            except ValueError:
                val = self.scripts.pop(key)
                logger.warning("%s option %s in config file not understood, set to False", key, val)
                val = False
        else:
            val = False
        return val

    def available_commands(self):
        """Sends a list with the available commands"""
        msg_str = "Available commands: " + ", ".join(self.script_names)
        self.send_msg(msg_str)

    @classmethod
    def configure_me(cls):
        print("")
        print(" # Scripts module ")
        print("This is the configuration helper for the scripts module, leave empty to exit")
        dict_out = {cls.key_name: {}}
        while True:
            script_command = input(" Introduce the command name: ")
            sc_cmd = script_command.strip()
            if not sc_cmd:
                break
            script_file = input(f" Introduce the path of the command to run with '{sc_cmd}': ")
            dict_out[cls.key_name][sc_cmd] = script_file
        return dict_out

    def telegram_message(self, msg):
        if not self.check_identity(msg) and not self._allow_everyone:
            self.blocked = True
            self.send_msg("You are not allowed to run scripts here")
        if self.blocked:
            return

        if not msg.has_arguments:
            self.send_msg("Add a command name after /script")
            return

        script_name, script_args = _parse_cmd_args(msg.text.strip(), self._run_in_shell)
        if script_name == "list":
            self.available_commands()
        elif script_name not in self.script_names:
            self.send_msg(f"Command {script_name} not defined")
            self.available_commands()
        else:
            command_path = pathlib.Path(self.scripts[script_name])
            logger.info("Running %s, path: %s", script_name, command_path)
            if script_args:
                logger.info("With args: %s", script_args)
            if command_path.is_file():
                try:
                    if self._run_in_shell:
                        full_cmd = f"{command_path} {' '.join(script_args)}"
                        sp.run(full_cmd, shell=True, check=True, cwd=command_path.parent)
                    else:
                        cmd_list = [f"./{command_path.name}"] + script_args
                        sp.run(cmd_list, check=True, cwd=command_path.parent)
                    self.send_msg("Command ran")
                except sp.CalledProcessError:
                    self.send_msg("Command ran but failed")
            else:
                self.send_msg("Script not found")



"""
    This module contains a mapping between Telegram commands
    (msgs which start with / ) and components which will act on them.

    Note that the components are only imported when/if the appropriate command is invoked
    This is a design choice as this way it is not necessary to have all dependencies
    if you want to run only some submodules of the pybliotecario.
"""
import importlib
import logging

log = logging.getLogger(__name__)


def import_component(module, cls):
    module_path = "pybliotecario.components." + module
    module = importlib.import_module(module_path)
    return getattr(module, cls)


def send_help(tele_api, chat_id):
    log.info("Sending help msg")
    components = [
        ("pid", "ControllerPID"),
        ("ip_lookup", "IpLookup"),
        ("arxiv_mod", "Arxiv"),
        ("scripts", "Script"),
        ("dnd", "DnD"),
        ("reactions", "Reactions"),
        ("wiki", "WikiComponent"),
        ("system", "System"),
        ("stocks", "Stocks"),
        ("twitter", "TwitterComponent"),
        ("photocol", "PhotoCol"),
    ]
    full_help = []
    for module, cls in components:
        try:
            Actor = import_component(module, cls)
            full_help.append(Actor.help_msg())
        except ModuleNotFoundError:
            pass
    tele_api.send_message("\n".join(full_help), chat_id)
    return None


def act_on_telegram_command(tele_api, message_obj, config):
    """
    Act for a given telegram command
    """
    tg_command = message_obj.command
    chat_id = config["DEFAULT"]["chat_id"]

    try:
        if tg_command == "ip":
            from pybliotecario.components.ip_lookup import IpLookup as Actor
        elif tg_command.lower() in ("is_pid_alive", "kill_pid"):
            from pybliotecario.components.pid import ControllerPID as Actor
        elif tg_command in ("arxiv_query", "arxiv", "arxivget", "arxiv_get"):
            from pybliotecario.components.arxiv_mod import Arxiv as Actor
        elif tg_command == "script":
            from pybliotecario.components.scripts import Script as Actor
        elif tg_command in ("r", "roll"):
            from pybliotecario.components.dnd import DnD as Actor
        elif tg_command in ("reaction_save", "reaction", "reaction_list"):
            from pybliotecario.components.reactions import Reactions as Actor
        elif tg_command[:4] == "wiki":
            from pybliotecario.components.wiki import WikiComponent as Actor
        elif tg_command == "system":
            from pybliotecario.components.system import System as Actor
        elif tg_command == "stock_price":
            from pybliotecario.components.stocks import Stocks as Actor
        elif tg_command.startswith("twitter_"):
            from pybliotecario.components.twitter import TwitterComponent as Actor
        elif tg_command.startswith("photocol"):
            from pybliotecario.components.photocol import PhotoCol as Actor

        elif tg_command == "help":
            return send_help(tele_api, chat_id)
        else:
            log.info(f"No actor declared for this command: {tg_command}")
            return None

    except ModuleNotFoundError as e:
        log.error(f"The component {tg_command} raised the following error: {e}")
        return tele_api.send_message(f"Dependencies for {tg_command} are missing", chat_id)

    actor_instance = Actor(
        tele_api,
        chat_id=chat_id,
        configuration=config,
        interaction_chat=message_obj.chat_id,
        running_in_loop=True,
    )
    return actor_instance.telegram_message(message_obj)



"""
    This module contains utilities for interacting with processes of the computer
    in which the bot runs.
    It allows for things like killing a process from Telegram or querying for
    an active process
"""

import logging

import psutil

from pybliotecario.components.component_core import Component

logger = logging.getLogger(__name__)


def get_process(pid):
    """Returns a process object for the given PID"""
    exists = psutil.pid_exists(pid)
    proc = None
    if exists:
        try:
            proc = psutil.Process(pid=pid)
        except psutil.NoSuchProcess:
            pass

    if proc is None:
        logger.warning("Process %s was not found", pid)

    return proc


def wait_for_it_until_finished(pids):
    """Receives a list of PIDs and wait for them"""
    processes = []
    for pid in pids:
        proc = get_process(pid)
        if proc is not None:
            processes.append(proc)
    psutil.wait_procs(processes)


def is_it_alive(data):
    """Given a pid or a string, check whether
    there is any matching process alive"""
    alive = False
    matches = []
    if data.isdigit():
        alive = psutil.pid_exists(int(data))
    else:
        all_active_processes = psutil.process_iter()
        for proc in all_active_processes:
            cmdline = proc.cmdline()
            for info in cmdline:
                if data in info:
                    command_str = " ".join(cmdline)
                    matches.append(f"PID: {proc.pid}, {command_str}")
                    alive = True
    if alive:
        msg = f"{data} is alive"
        if matches:
            matching_processes = "\n > ".join(matches)
            msg += f"\nI found the following matching processes: \n > {matching_processes}"

    else:
        msg = f"{data} not found among active processes"
    return msg


def kill_pid(pid):
    """Kills the given pid"""
    process = get_process(pid)
    if process is not None:
        process.kill()
        return f"{pid} killed"
    else:
        return f"No process with pid {pid}"


class ControllerPID(Component):
    """Check or kill pids"""

    help_text = """ > PID module
    /kill_pid pid: kills a given pid
    /is_pid_alive pid/name_of_program: looks for the given pid or program to check whether it is still alive"""

    def cmdline_command(self, args):
        """Waits until the given PID(s) are finished"""
        logger.info("Waiting for the given PIDs: %s", args.pid)
        wait_for_it_until_finished(args.pid)

    @staticmethod
    def kill(pid):
        """Kills the received PID"""
        return kill_pid(pid)

    @staticmethod
    def alive(pid):
        """Check whether a PID (or str for searching
        for a PID) is alive"""
        return is_it_alive(pid)

    def telegram_message(self, msg):
        if self.check_identity(msg):
            pid_string = msg.text.strip()
            if msg.command == "kill_pid":
                if pid_string.isdigit():
                    return_msg = self.kill(int(pid_string))
                else:
                    return_msg = f"{pid_string} is not a PID?"
            elif msg.command == "is_pid_alive":
                return_msg = self.alive(pid_string)
            else:
                return_msg = f"Command {msg.command} not understood?"
        else:
            return_msg = "You are not allowed to use this"
        self.send_msg(return_msg)



"""
    Mock backend in order to test the pybliotecario
    without communication with any service
"""

from datetime import datetime
import pathlib

from pybliotecario.backend.telegram_util import TelegramMessage

TESTID = 1234  # chat id for the test backend
_TESTUSER = "hiro"


class TestMessage(TelegramMessage):
    """Copy of the TelegramMessage class"""

    _type = "Test"


def _create_fake_msg(text):
    """Create fake messages for testing"""
    import numpy as np

    ret = {
        "update_id": np.random.randint(1000),
        "message": {
            "message_id": np.random.randint(1000),
            "from": {
                "id": TESTID,
                "is_bot": False,
                "first_name": _TESTUSER,
                "last_name": _TESTUSER,
                "username": _TESTUSER,
                "language_code": "en",
            },
            "chat": {
                "id": TESTID,
                "first_name": _TESTUSER,
                "last_name": _TESTUSER,
                "username": _TESTUSER,
                "type": "private",
            },
            "date": datetime.now().toordinal(),
            "text": text,
        },
    }
    return ret


class TestUtil:
    """
    The test utility tries to behave as other backend would

    It provides _get_updates / send_message just as any other backend,
    but it writes/reads the update from a communication file.

    If fake_msgs are given, it will read the messages from those instead
    of from the communication_file
    Likewise, if a communication_file is not given, it will return the message
    instead of writing it down

    Parameters
    ----------
        communication_file: str/pathlib
            file to write/read from
        fake_msg: list(str)
            list of messages to generate
    """

    def __init__(self, communication_file=None, fake_msgs=None, **kwargs):
        if communication_file is None:
            self.comm_file = False
        else:
            self.comm_file = pathlib.Path(communication_file)
        self.fake_msgs = fake_msgs

    def _get_updates(self, not_empty=False):
        """
        Returns a json with the latest messages the bot has received

        This test function creates a list of two mock messages which
        will be in turn parsed by the Message class
        """
        if self.fake_msgs is not None:
            msgs = self.fake_msgs
        elif self.comm_file:
            msgs = self.comm_file.read_text().split("\n")
        else:
            msgs = ["This is only a test", "Another one"]
        return [_create_fake_msg(i) for i in msgs]

    def act_on_updates(self, action_function, not_empty=False):
        """
        Receive the input using _get_updates, parse it with
        the telegram message class and act in consequence
        """
        all_updates = self._get_updates(not_empty=not_empty)
        for update in all_updates:
            msg = TestMessage(update)
            action_function(msg)

    def send_message(self, text, chat, **keywords):
        """
        Sends a message to the communication_file this class has been
        instantiated with
        """
        if self.comm_file:
            with open(self.comm_file, "a") as f:
                f.write(f"{text}\n")
            return None
        return text

    def send_image(self, *args):
        """Writes the img_path to the comm file"""
        return self.send_message(*args)

    def send_file(self, *args):
        """Writes the file_path to the comm file"""
        return self.send_message(*args)

    # Auxiliary
    def is_msg_in_file(self, msg):
        """Check whether the given message is in the comms file
        This is something that is only useful for the TestUtil backend"""
        read_text = self.comm_file.read_text()
        return msg in read_text



"""
    Test the system component
"""
import subprocess as sp
from pybliotecario.pybliotecario import main
from pybliotecario.backend import TestUtil
from .conftest import generate_fake_config


def test_uptime(tmp_path, tmpfile):
    """ Checks that the uptime is correct """
    fake_config = generate_fake_config(tmp_path)
    fake_msg = ["/system uptime"]
    test_util = TestUtil(communication_file=tmpfile, fake_msgs=fake_msg)
    args = ["-d", "--exit_on_msg"]
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    # Get the uptime from system
    result_cmd = sp.run("uptime", capture_output=True, check=True)
    result = result_cmd.stdout.decode().strip()
    # Get the telegram time
    message = tmpfile.read_text().strip()
    assert result == message



"""
    Module to save images into an image collection
    with thumbnails and keeping a json as database

    It's a companion to https://github.com/scarlehoff/websito/blob/master/views/foto.pug
"""
from datetime import datetime
import json
import logging
from pathlib import Path
import uuid

from PIL import Image

from pybliotecario.components.component_core import Component

log = logging.getLogger(__name__)
CONFIG = "PHOTOCOL"
ACCEPTED_COMMANDS = ["photocol", "photocol_remove"]


class PhotoCol(Component):
    """
    Save pictures to the photocol folder
    and to the photocol.json database from telegram
    """

    help_text = """ > PhotoCol module
    /photocol comment to add to the picture
    /photocol_remove remove a picture from the db given the unique identifier
    """

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        photocol_config = self.read_config_section(CONFIG)
        self._photofol = Path(photocol_config.get("folder"))
        self._photodb = Path(photocol_config.get("db"))

    def _save_picture(self, file_id):
        """Save the picture to the target folder and creates a thumbnail"""
        unique_name = str(uuid.uuid4())
        file_path = (self._photofol / unique_name).with_suffix(".jpg")
        self.telegram.download_file(file_id, file_path)

        # Create also a thumbnail
        try:
            image = Image.open(file_path)
            aratio = image.width / image.height
            # Keep a resemblance of the original aspect ratio
            size = 240
            thumb_size = (size, int(size / aratio))
            image.thumbnail(thumb_size)
            image.save((self._photofol / "thumbnail" / unique_name).with_suffix(".jpg"))
        except IOError:
            return None

        return file_path

    def _update_db(self, photo_path, comment):
        """Adds a new entry to the json database with the date of (today)
        the path to the picture and a comment
        """
        # Read the previous json if any
        foto_list = []
        if self._photodb.exists():
            foto_list = json.load(self._photodb.open("r", encoding="utf-8"))

        today = datetime.today()
        foto_list.insert(
            0, {"photo": photo_path.name, "date": today.strftime("%d/%m/%Y"), "comment": comment}
        )
        json.dump(foto_list, self._photodb.open("w", encoding="utf-8"), indent=True)

    def _remove_from_db(self, uuid):
        """Remove an entry from the db"""
        test = f"{uuid}.jpg"
        if not self._photodb.exists():
            return self.send_msg("There's no database to remove anything from!")

        foto_list = json.load(self._photodb.open("r", encoding="utf-8"))
        new_json = []
        success = False
        for foto in foto_list:
            if test != foto["photo"]:
                success = True
                new_json.append(foto)

        if success:
            # Create a removed folder to put the picture in
            removed_folder = self._photofol / "removed"
            removed_folder.mkdir(exist_ok=True)

            # Move both picture and thumbnail to removed
            foto_path = self._photofol / test
            foto_thumb = self._photofol / "thumbnail" / test
            foto_path.rename(removed_folder / test)
            foto_thumb.rename(removed_folder / f"{test}-thumb")

            json.dump(new_json, self._photodb.open("w", encoding="utf-8"), indent=True)
            return self.send_msg("Entry removed!")
        else:
            return self.send_msg("That entry was not found in the json db")

    def telegram_message(self, msg):
        if msg.command not in ACCEPTED_COMMANDS:
            return self.send_msg("Command not understood")

        if not self.check_identity(msg):
            self.send_msg("You are not allowed to interact with this command!")
            return

        if msg.command == "photocol_remove":
            return self._remove_from_db(msg.text)

        # Save picture
        photo_path = self._save_picture(msg.file_id)

        if photo_path is None:
            return self.send_msg("There was a problem when creating the thumbnail for this picture")
        # Update json
        self._update_db(photo_path, msg.text)
        self.send_msg("Picture saved")



"""
    Define custom parsers for the config reader
    and default data/config locations
"""
from configparser import ConfigParser
from copy import copy
from os import environ
from pathlib import Path

NAME = "pybliotecario"


def default_config_path(name=NAME):
    """Return the default config path looking at the value of XDG_CONFIG_HOME
    Usually: $HOME/.config/pybliotecario/pybliotecario.ini
    """
    config_folder = Path(environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
    return config_folder / name / "pybliotecario.ini"


def default_data_path(name=NAME):
    """Return the default data path looking at the value of XDG_DATA_HOME
    Usually: $HOME/.local/share/pybliotecario
    """
    data_folder = Path(environ.get("XDG_DATA_HOME", Path.home() / ".local" / "share"))
    return data_folder / name


def _parse_chat_id(value):
    """The chat id can come as a list or as a single value,
    this ensures that it is always read as a list"""
    split_values = value.split(",")
    return [int(i.strip()) for i in split_values]


def _parse_main_chat_id(value):
    """Regardless of the format of the chat id (list or single value)
    return the main id (i.e., the only one there is or the first one
    """
    return _parse_chat_id(value)[0]


class CustomConfigParser(ConfigParser):
    """Equal to ConfigParser with a number of predefined converters"""

    def __init__(self, *args, converters={}, allow_no_value=True, **kwargs):
        new_converters = copy(converters)
        new_converters.setdefault("idlist", _parse_chat_id)
        new_converters.setdefault("mainid", _parse_main_chat_id)
        super().__init__(*args, converters=new_converters, allow_no_value=allow_no_value, **kwargs)



"""
    This module contains a mapping between the command line arguments
    and the components which run them.

    Note that the components are only imported when/if the appropriate command is invoked
    This is a design choice as this way it is not necessary to have all dependencies
    if you want to run only some submodules of the pybliotecario.

    The most obvious example is running the program just to send msgs and files to Telegram
"""

import logging

log = logging.getLogger(__name__)


def run_command(args, tele_api, config):
    """
    Receives the whole batch of arguments and acts accordingly
    """
    chat_id = args.chat_id
    if not chat_id.isnumeric():
        # If the id is not numeric maybe is an alias that we have
        try:
            chat_id = config["ALIAS"][chat_id]
        except KeyError:
            chat_id = ""
    # If after everything, chat_id is still empty:
    if not chat_id:
        chat_id = config["DEFAULT"]["chat_id"]

    actors = []

    if args.my_ip:
        from pybliotecario.components.ip_lookup import IpLookup

        actors.append(IpLookup)

    if args.pid:
        from pybliotecario.components.pid import ControllerPID

        actors.append(ControllerPID)

    if args.weather:
        from pybliotecario.components.weather import Weather

        actors.append(Weather)

    if args.check_repository:
        from pybliotecario.components.repositories import Repository

        actors.append(Repository)

    if args.check_github_issues:
        from pybliotecario.components.github_component import Github

        actors.append(Github)

    if args.arxiv_new:
        from pybliotecario.components.arxiv_mod import Arxiv

        actors.append(Arxiv)

    if args.stock_watcher is not None:
        # stock_watcher could be an empty list!
        from pybliotecario.components.stocks import Stocks

        actors.append(Stocks)

    for Actor in actors:
        actor_instance = Actor(tele_api, chat_id=chat_id, configuration=config)
        actor_instance.cmdline_command(args)

    # These three are the basic commands:
    # send file, send image, send text
    if args.file:
        tele_api.send_file(args.file, chat_id)
        log.info("File sent")

    if args.image:
        tele_api.send_image(args.image, chat_id)
        log.info("Image sent")

    if args.message:
        message_text = " ".join(args.message)
        # Unescape certain characters that we know to be fine
        message_text = message_text.replace("\\n", "\n")
        tele_api.send_message(message_text, chat_id)
        log.info("Message sent")



"""
    Wrapper around the GitHub API

    To get an API key go to 
    https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
    and follow the instructions
"""
import datetime
import logging

import github as pygithub

from pybliotecario.components.component_core import Component

log = logging.getLogger(__name__)


class Github(Component):
    """
    Checks what has happened in the given repository
    """

    key_name = "GITHUB"

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        self.github_config = self.read_config_section(self.key_name)
        self._token = self.github_config.get("token", "")
        self._hours = int(self.github_config.get("since_hours", 2))
        self._github = None

    @property
    def github(self):
        if self._github is None:
            self._github = pygithub.Github(self._token)
        return self._github

    @classmethod
    def configure_me(cls):
        print("")
        print(" # Github module ")
        print("This is the configuration helper for the github module")
        print("Please introduce your access token: ")
        access_token = input(" > ")
        print("Show only the issues opened in the last X hours:")
        hours = None
        while hours is None:
            try:
                hours = int(input(" > "))
            except ValueError:
                print("Only integer numbers accepted")
        dict_out = {cls.key_name: {"token": access_token, "since_hours": hours}}
        return dict_out

    def _check_issues(self, repo_name):
        """Get the issues in the last two hours"""
        repo = self.github.get_repo(repo_name)
        since_datetime = datetime.datetime.now() - datetime.timedelta(hours=self._hours)
        all_issues = repo.get_issues(since=since_datetime)
        titles = []
        for issue in all_issues:
            if issue.created_at < since_datetime:
                continue
            pr = ""
            if issue.pull_request is not None:
                pr = "PR"
            url = f"https://github.com/{repo_name}/issues/{issue.number}"
            titles.append(f"    > {pr}#{issue.number}: [{issue.title}]({url})")
        if titles:
            self.send_msg(
                f"New github issues/PR in {repo_name}: \n" + "\n".join(titles), markdown=True
            )

    def cmdline_command(self, args):
        repository = args.check_github_issues
        self._check_issues(repository)


if __name__ == "__main__":
    import configparser
    from pathlib import Path
    import tempfile

    from pybliotecario.backend import TestUtil
    from pybliotecario.pybliotecario import logger_setup, main

    tmpdir = Path(tempfile.mkdtemp())
    tmptxt = tmpdir / "text.txt"

    config = configparser.ConfigParser()
    config["DEFAULT"] = {"main_folder": tmpdir, "token": "AAAaaa123", "chat_id": 453}
    config["GITHUB"] = {"token": "<TOKEN GITHUB>", "since_hours": 5}

    logger_setup(tmpdir / "log.txt", debug=True)

    log.info("Testing the github component")
    test_util = TestUtil(communication_file=tmptxt)
    args = ["--check_github_issues", "NNPDF/nnpdf"]
    main(cmdline_arg=args, tele_api=test_util, config=config)
    messages = tmptxt.read_text()
    print("Results: ")
    print(messages)



"""
    Tests the PID module
"""
import time
import subprocess as sp
from multiprocessing.pool import ThreadPool as Pool
from pybliotecario.pybliotecario import main
from pybliotecario.backend import TestUtil
from .conftest import generate_fake_config


def test_cmdline_usage(tmpfile):
    """Tests that upon --pid PID
    the pybliotecario waits until the given PID is finished
    """
    # Send a sleep command to the background
    sleep_cmd = sp.Popen(["sleep", "600s"])
    # Now send a pybliotecario to the background waiting for sleep to finish
    pool = Pool(processes=1)
    test_util = TestUtil(communication_file=tmpfile)
    msg = "Hola, caracola"
    kwargs = {
        "cmdline_arg": [msg, "--pid", str(sleep_cmd.pid), "--debug"],
        "config": generate_fake_config(tmpfile),
        "tele_api": test_util,
    }
    pool.apply_async(main, kwds=kwargs)
    # Check that nothing has yet been written
    if tmpfile.exists():
        assert tmpfile.read_text() == ""
    time.sleep(1)
    # Kill it on its sleep
    sleep_cmd.kill()
    # Wait a few seconds for the dust to settle
    time.sleep(1)
    # Now check that the message has been written
    assert tmpfile.exists()
    assert tmpfile.read_text().strip() == msg


def test_msg_usage_alive(tmpfile):
    """Tests that upon received the message "/is_pid_alive PID"
    responds correctly if it is alive"""
    # Send a sleep command to the background
    sleep_cmd = sp.Popen(["sleep", "600s"])

    fake_msg = [
        f"/is_pid_alive {sleep_cmd.pid}",  # The answer should be yes
        "/is_pid_alive 123456",  # The answer should be no
    ]
    test_util = TestUtil(communication_file=tmpfile, fake_msgs=fake_msg)
    args = ["-d", "--exit_on_msg"]
    fake_config = generate_fake_config(tmpfile)
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    # Check whether what we received is a valid IP
    message = tmpfile.read_text().strip().split("\n")
    assert len(message) == 2
    assert "alive" in message[0]
    assert "not found" in message[1]



"""
    Module to interact with twitter

    In order to use this module, create a new application in the
    [twitter developers portal](https://developer.twitter.com/)
    and fill in `pybliotecario.ini` the keys requested by `_authenticate`:

    ```ini
    [TWITTER]
    access_token = <acces_token>
    access_token_secret = <access_token_secret>
    consumer_key = <consumer_key>
    consumer_secret = <consumer_secret>
    ```
"""
import logging

import tweepy as tw

from pybliotecario.components.component_core import Component

logger = logging.getLogger(__name__)
COMP_NAME = "TWITTER"


def _authenticate(config):
    """Authenticate a user using a consumer key and consumer secret"""
    at = config["access_token"]
    ats = config["access_token_secret"]
    ck = config["consumer_key"]
    cs = config["consumer_secret"]
    auth = tw.OAuth1UserHandler(
        consumer_key=ck, consumer_secret=cs, access_token=at, access_token_secret=ats
    )
    return tw.API(auth)


def _parse_tweet(tweet):
    """Convert a tweet into a text string"""
    author = tweet.author
    text = tweet.text
    return f"{author.name} (@{author.screen_name}): {text}"


def _prepare_tweet_list(list_of_tweets):
    """Make a list of tweets into a string that can be sent"""
    all_tweets = []
    for tweet in list_of_tweets:
        all_tweets.append(_parse_tweet(tweet))
    return "\n-\n".join(all_tweets)


def _parse_tl_command(command_data):
    """Parse the command data from a TL
    This can be:
        @user N
        N @user
        @user
        N

    Returns
    -------
        user: username
        N: number of tweets to retrieve
    """
    if "@" in command_data:
        data = command_data.split()
        # If there are extra stuff we don't care
        if len(data) < 2:
            return data[0], None
        if data[0].startswith("@"):
            username = data[0][1:]
            N = data[1]
        else:
            username = data[1][1:]
            N = data[0]
        return username, N
    return None, command_data


class TwitterComponent(Component):
    """
    Interact with twitter
    """

    help_text = """
    > Twitter module
    /twitter_tl [@user] [N=20]: send the last N tweets from the TL [of user @user]
    /twitter_mentions [N=5]: send the last N mentions
    /twitter_tweet <text>: send a tweet"""

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        self._api = _authenticate(self.read_config_section(COMP_NAME))

    @classmethod
    def configure_me(cls):
        # TODO:
        # Add automatic get token thing
        pass

    def _get_timeline(self, n=None):
        """Send the last ``n`` tweets from the user timeline"""
        if n is None:
            n = 20
        return _prepare_tweet_list(self._api.home_timeline(count=n))

    def _get_timeline_from_user(self, user, n=None):
        """Send the last ``n`` tweets from @user"""
        if n is None:
            n = 10
        return _prepare_tweet_list(self._api.user_timeline(screen_name=user, count=n))

    def _get_mentions(self, n=None):
        """Send the last ``n`` mentions of the user"""
        if n is None:
            n = 5
        return _prepare_tweet_list(self._api.mentions_timeline(count=n))

    def _send_tweet(self, tweet_msg):
        """Sends tweet_msg as a tweet"""
        if len(tweet_msg) > 280:
            self.send_msg("The tweet is too long! Please limit yourself to 280 characters!")
            return
        self._api.update_status(tweet_msg)
        return "tweet sent"

    def telegram_message(self, msg):
        """Digest the telegram msg"""
        command = msg.command
        data = msg.text.strip()
        if command == "twitter_tl":
            user, n = _parse_tl_command(data)
            if user is None:
                response = self._get_timeline(n)
            else:
                response = self._get_timeline_from_user(user, n)
        elif command == "twitter_mentions":
            response = self._get_mentions(data)
        elif command == "twitter_tweet":
            response = self._send_tweet(data)
        else:
            response = "Command not recognized"
        self.send_msg(response)



"""
    This component contains system-commands to be run
    remotely
"""
import subprocess as sp

from pybliotecario.components.component_core import Component


def run_command(command):
    """Run the given command using subprocess and return the stdout"""
    try:
        result = sp.run(command, capture_output=True, check=True)
        stdout = result.stdout.decode().strip()
    except FileNotFoundError:
        stdout = f"No such command: {command}"
    except sp.CalledProcessError:
        stdout = f"The command {command} failed"
    return stdout


ACCEPTED_COMMANDS = {"uptime": "uptime"}


class System(Component):
    help_text = """ > System component
    /system uptime: returns the uptime of the computer in which the bot lives"""

    def __init__(self, telegram_object, **kwargs):
        super().__init__(telegram_object, **kwargs)

    def telegram_message(self, msg):
        # Allow only the main user to use system
        if not self.check_identity(msg):
            self.send_msg("You are not allowed to run scripts here")
            return
        command_key = msg.text.strip()
        command_name = ACCEPTED_COMMANDS.get(command_key)
        # Check whether the command is in the accepted_commands dictionary
        if command_name is None:
            self.send_msg(f"Command {command_name} not recognized")
        else:
            self.send_msg(run_command(command_name))



"""
    Wrapper for the argument parser and the initialization
"""
from argparse import Action, ArgumentParser, ArgumentTypeError
import configparser
import glob
import importlib
import os
from pathlib import Path

from . import components
from .backend import TelegramUtil
from .backend.telegram_util import TelegramMessage
from .components.component_core import Component
from .customconf import default_config_path, default_data_path


def validpath(value):
    """Check whether the received path is valid"""
    path = Path(value)
    if not path.exists():
        raise ArgumentTypeError(f"The file '{value}' can't be found")
    if path.is_dir():
        raise ArgumentTypeError(f"'{value}' is a directory, only single files can be sent")
    return path


def write_config(config_dict, config_file, config_exists=False):
    """
    Receives a dictionary of dictionaries and makes it into a
    ConfigParser
    """
    config = configparser.ConfigParser()
    if config_exists:
        config.read(config_file)
    for key, item in config_dict.items():
        config[key] = item
    with open(config_file, "w") as f:
        config.write(f)


def check_attr(attr):
    """
    Checks whether the attr is a class and whether
    it does inherit from Component and whether
    """
    if not isinstance(attr, type):
        return False
    if attr == Component:
        return False
    if issubclass(attr, Component):
        # Is there a more elegant way of knowing whether a method is overriden?
        return not Component.configure_me.__code__ == attr.configure_me.__code__
    return False


def config_module(module):
    """
    Calls the .configure_me method of all modules which
    inherit from component_core.Component
    """
    actor_list = []
    for name in dir(module):
        name_py = getattr(module, name)
        if check_attr(name_py):
            actor_list.append(name_py)
    dict_list = []
    for Actor in actor_list:
        name = Actor.whoamI()
        yn = input(f"Do you want to configure {name} now? [yn] ")
        if not yn.lower().startswith(("y", "s")):
            continue
        result = Actor.configure_me()
        if result:
            dict_list.append(result)
    return dict_list


def configure_telegram():
    """Configure pybliotecario to use the Telegram API
    This function walks the user through the process of creating a new bot
    and storing the token and the chat_id of the user in the configuration file
    """
    # Initialize the bot in telegram
    print(
        """Welcome to The Wizard!
The first thing you will need is to create a new bot with the botfather in your telegram client.
Alternatively, you can get use a previously created bot.
Ask the botfather for an authorization token before continuing.
The instructions on how to get the token can be found here: https://core.telegram.org/bots#how-do-i-create-a-bot
"""
    )
    token = input("Authorization token: ")

    # Try to fire up the bot with the given token
    telegram_API = TelegramUtil(token, timeout=20)
    print("Thanks, let's test this out. Say something (anything!) to your bot in telegram")

    for _ in range(20):  # Allow for 20 tries
        all_updates = telegram_API.raw_updates()

        for update in all_updates:
            msg = TelegramMessage(update)
            print(f"Message received: {msg.text}")
            yn = input("Was this your msg? [y/n] ")
            if msg_found := yn.lower().startswith(("y", "s")):
                chat_id = msg.chat_id
                username = msg.username
                print(f"Your chat id is: {chat_id} and your username: {username}")
                break

        if msg_found:
            break

        print("Please, try again")

    else:
        raise ValueError("There was some problem with the configuration of Telegram")

    yn = input(
        "Do you want to enable the 'chivato' mode, so you get a warning if anyone other than you tries to use the bot? [yn] "
    )
    chivato = yn.lower().startswith(("y", "s"))

    # Now prepare the dictionary with the DEFAULT field of the config file
    config_dict = {"DEFAULT": {"TOKEN": token, "chat_id": chat_id, "chivato": chivato}}
    return config_dict


def configure_all():
    """Import everything inside the components folder that
    inherint from Component and run config_module
    on it"""
    config_dict = {}
    missing_dependencies = False

    folder_components = os.path.dirname(components.__file__)
    module_components = components.__name__
    modules = glob.glob(folder_components + "/*.py")
    for module_file in modules:
        module_name = Path(module_file).with_suffix("").name
        try:
            module = importlib.import_module(f"{module_components}.{module_name}")
        except ModuleNotFoundError:
            print(
                f"In order to use the component '{module_name}' it is necessary to install its dependencies"
            )
            missing_dependencies = True
            continue
        dict_list = config_module(module)
        for dictionary in dict_list:
            for key, item in dictionary.items():
                config_dict[key] = item

    if missing_dependencies:
        print(
            """To install missing dependencies for a particular component you can install them explicitly:
    ~$ pip install pybliotecario[component]
or install all dependencies with
    ~$ pip install pybliotecario[full]"""
        )

    return config_dict


class InitAction(Action):
    """
    This class performs the initialization.

    Everything is only imported exactly where it is needed because that way
    also serves as a sort of documentation? I think. It looks instructive...
    """

    def __init__(self, nargs=0, **kwargs):
        super().__init__(nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, *args, **kwargs):
        """
        Configures the pybliotecario by first configuring the Telegram API
        and then calling the ``configure_me`` method of all components
        """
        config_dict = {}

        config_path = default_config_path()
        data_folder = default_data_path()

        if namespace.config_file is not None:
            print(
                f"WARNING! You are setting {namespace.config_file} as the configuration file instead of the default {config_path}"
            )
            yn = input("Are you sure? Do you want to continue [y/n] ")
            if not yn.lower().startswith(("y", "s")):
                print("Don't use --config_file with --init to avoid this!")
                parser.exit(0)
            config_path = namespace.config_file
            data_folder = Path(input(f"Where do you want your data? e.g.: {data_folder}: "))

        print(f"Note: configuration will be written to {config_path}")

        # Make sure the folders exist
        data_folder.mkdir(exist_ok=True, parents=True)
        config_path.parent.mkdir(exist_ok=True, parents=True)

        # Check whether a config file already exists, if it does, ask before doing the Telegram
        initialize = True
        if config_exists := config_path.exists():
            print("It seems pybliotecario has already been initialized in this computer")
            yn = input("Do you want to start the configuration from scratch? [y/n] ")
            initialize = yn.lower().startswith(("y", "s"))

        if initialize:
            config_dict.update(configure_telegram())
            config_dict["DEFAULT"]["main_folder"] = data_folder

        print("Now let's loop over all pybliotecario modules to configure their options")
        config_dict.update(configure_all())
        # And finally write the config file
        write_config(config_dict, config_path, config_exists=config_exists)
        print(f"The configuration of the pybliotecario has been written to: {config_path}")
        parser.exit(0)


def parse_args(args):
    """Wrapper for ArgumentParser"""
    parser = ArgumentParser()
    parser.add_argument(
        "--init", help="Wizard to configure the pybliotecario for the first time", action=InitAction
    )
    parser.add_argument(
        "--config_file",
        help=f"Define a custom configuration file (default: {default_config_path()})",
    )
    parser.add_argument(
        "--backend",
        help="Choose backend: telegram (default), facebook",
        type=str,
        default="Telegram",
    )

    parser_cmd = parser.add_argument_group("Command line program")
    parser_cmd.add_argument("message", help="Message to send to Telegram", nargs="*")
    parser_cmd.add_argument("-i", "--image", help="Send image to Telegram")
    parser_cmd.add_argument("-f", "--file", help="Send file to Telegram", type=validpath)
    parser_cmd.add_argument("--chat_id", help="Chat id to send the message to", default="")

    parser_dae = parser.add_argument_group("Pybliotecario daemon")
    parser_dae.add_argument("-d", "--daemon", help="Activate the librarian", action="store_true")
    parser_dae.add_argument(
        "--debug",
        help="Write everything to terminal instead of writing to log file",
        action="store_true",
    )
    parser_dae.add_argument(
        "--clear_incoming",
        help="Clears incoming messages in case something has gone wrong",
        action="store_true",
    )
    parser_dae.add_argument(
        "--exit_on_msg",
        help="Exit after receiving the first batch of messages",
        action="store_true",
    )

    parser_com = parser.add_argument_group("Pybliotecario's components")
    parser_com.add_argument(
        "--arxiv_new",
        help="Send a msg containing a digest of the new submissions to arxiv",
        action="store_true",
    )
    parser_com.add_argument(
        "--weather",
        help="Sends a msg to telegram with the current weather and a small forecast",
        action="store_true",
    )
    parser_com.add_argument(
        "--check_repository",
        help="Sends a msg to telegram with the incoming information for the given repository",
    )
    parser_com.add_argument("--check_github_issues", help="Get the latest github issues")
    parser_com.add_argument(
        "--pid",
        help="Monitor a PID and sends a message when the PID is finished",
        type=int,
        nargs="+",
    )
    parser_com.add_argument(
        "--my_ip",
        help="Send to the default chat the current IP of the computer",
        action="store_true",
    )
    parser_com.add_argument(
        "--stock_watcher",
        help="Looks at json file (can be given in config) to watch a number of stocks",
        nargs="*",
    )
    return parser.parse_args(args)



"""
    Use yahoo_fin: http://theautomatic.net/yahoo_fin-documentation/
    to obtain information of stock prices

    The watch list is a json like:
    {
    "AAPL" : {
        "below" : 100,
        "above" : 170
    },

"""
import json
import logging

from pybliotecario.components.component_core import Component

logger = logging.getLogger(__name__)


def _get_live_price(ticker):
    """Wrapper around stock_info.get_live_price to avoid
    importing yahoo_fin if not explicitly called"""
    from yahoo_fin import stock_info

    return stock_info.get_live_price(ticker)


### All checks to be applied, take ticker name, current price and conditions dictionary
def below(price, ticker, conditions):
    """Check whether the price is below the threshold"""
    thrs = conditions["below"]
    if price < thrs:
        return f"{ticker} price is below the threshold: {price:.3f} < {thrs}"
    return None


def above(price, ticker, conditions):
    """Check whether the price is below the threshold"""
    thrs = conditions["above"]
    if price > thrs:
        return f"{ticker} price is above the threshold: {price:.3f} > {thrs}"
    return None


CHECKS = [below, above]


def checking_conditions(price, ticker, conditions):
    """
    Check the price against the given conditions to decide on the msg to send (if any)
    """
    for check in CHECKS:
        if check.__name__ in conditions:
            ret = check(price, ticker, conditions)
            if ret is not None:
                # The first check that passes is sent
                return ret
    return None


def check_stock(json_file):
    """
    Reads up a json file given in the STOCKS::watch_json variable of the config file
    and sends a notification if any of the conditions occur
    """
    with open(json_file, "r") as f:
        information = json.load(f)
    results = []
    for ticker, conditions in information.items():
        # noinspection PyPep8
        try:
            current_price = _get_live_price(ticker)
            ret = checking_conditions(current_price, ticker, conditions)
            if ret is not None:
                results.append(ret)
        except AssertionError:
            logger.warning(f"ticker {ticker} not found")
        except:  # yahoo-fin is not the most stable piece of software
            logger.warning(f"There was a non-expected error with {ticker}")
    return results


class Stocks(Component):
    key_name = "STOCKS"
    help_text = """ > Stocks module
    /stock_value ticker: get the most recent information about the given ticker """

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        if self.key_name in configuration:
            self.stock_config = self.read_config_section("STOCKS")
        else:
            self.stock_config = None

    def cmdline_command(self, args):
        """`stock_watcher`: opens the given file and "follows the rules" """
        if args.stock_watcher is not None:
            if len(args.stock_watcher) == 0 and self.stock_config is not None:
                json_files = [self.stock_config.get("watch_json")]
            else:
                json_files = args.stock_watcher
            for ff in json_files:
                msgs = check_stock(ff)
                for msg in msgs:
                    self.send_msg(msg)

    def telegram_message(self, msg):
        command = msg.command
        ticker = msg.text.strip()
        if command == "stock_price":
            # noinspection PyPep8
            try:
                pp = _get_live_price(ticker)
                self.send_msg(f"{ticker} price: {pp:.3f}$")
            except:  # yahoo-fin not very stable :(
                self.send_msg(f"Unknown error trying to get information from {ticker}")


if __name__ == "__main__":
    import os
    import tempfile

    json_example = {"AAPL": {"below": 140, "above": 110}, "TSLA": {"below": 1000, "above": 200}}
    fd, tmp = tempfile.mkstemp(suffix=".json", text=True)
    with os.fdopen(fd, "w") as f:
        json.dump(json_example, f)
    msgs = check_stock(tmp)
    for i, msg in enumerate(msgs):
        print(f"Message #{i + 1}: {msg}")



"""
    Tests for the DnD component
"""

import re
from pybliotecario.pybliotecario import main
from pybliotecario.backend import TestUtil
from pybliotecario.components import dnd
from .conftest import generate_fake_config


def test_parse_roll():
    """ Test that the dice is parsed correctly """

    def check_comb(comb):
        dice, signs, modifiers = dnd.parse_roll(comb)
        total = ""
        for die, sign in zip(dice, signs):
            total += "+" if sign == 1 else "-"
            total += die
        if total[0] == "+":
            total = total[1:]
        assert f"{total}{modifiers}" == comb

    trials = ["1d20", "3d20+5d10-2d13+8"]
    for trial in trials:
        check_comb(trial)


def test_msg_roll(tmpfile):
    """
    Test that the rolled number are in the
    requested range
    """
    fake_msg = ["/r 1d20", "/r 4d20+4", "/r 1d20-1d12+12-1"]
    test_util = TestUtil(communication_file=tmpfile, fake_msgs=fake_msg)
    args = ["-d", "--exit_on_msg"]
    fake_config = generate_fake_config(tmpfile)
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    messages = tmpfile.read_text().strip()
    results = re.findall(r"(?<==\s)\d*", messages)
    assert 0 < int(results[0]) < 21
    assert 7 < int(results[1]) < 85
    assert -1 < int(results[2]) < 31



"""
    Tests the ip_lookup component
"""

import socket
from pybliotecario.pybliotecario import main
from pybliotecario.backend import TestUtil
from .conftest import generate_fake_config


def test_cmdline_usage(tmpfile):
    """Tests that upon --my_ip
    the pybliotecario does send the ip
    """
    test_util = TestUtil(communication_file=tmpfile)
    args = ["--my_ip"]
    fake_config = generate_fake_config(tmpfile)
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    # Check whether what we received is a valid IP
    message = tmpfile.read_text().strip()
    socket.inet_aton(message)


def test_msg_usage(tmpfile):
    """Tests that upon received the message "/ip"
    responds with something that looks like an IP"""
    fake_msg = ["/ip"]
    test_util = TestUtil(communication_file=tmpfile, fake_msgs=fake_msg)
    args = ["-d", "--exit_on_msg"]
    fake_config = generate_fake_config(tmpfile)
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    # Check whether what we received is a valid IP
    message = tmpfile.read_text().strip()
    socket.inet_aton(message)



#!/usr/bin/env python3
"""
    Facebook backend

    Using this backend will stat a flask server in the selected port.
    Testing this backend is a bit of a pain as one has to be in a server
    which facebook should be able to access with a valid SSL certificate.

    For quick testing, I am using the following setup:
    ~$ iptables -A INPUT -p tcp --dport 3000  -j ACCEPT
    ~$ ngrok http <my_personal_server>:3000

    And then I open the flask server in the port 3000 and give facebook
    the ngrok url.
    For actual deployment one would want to set up some actual server.
"""

import json
import logging
import pathlib

import requests

from pybliotecario.backend.basic_backend import Backend, Message

_HAS_FLASK = True
try:
    from flask import Flask, request
except ModuleNotFoundError:
    # Don't raise the error until this is not actually used
    _HAS_FLASK = False

logger = logging.getLogger(__name__)

FB_API = "https://graph.facebook.com/v2.12/me/messages"
MAX_SIZE = 2000


class FacebookMessage(Message):
    """Facebook implementation's of the Message class"""

    _type = "facebook"
    _group_info = None

    def _parse_update(self, update):
        """Receives an update in the form of a dictionary (that came from a json)
        and fills the message attributes
        """
        # Check whether it is the correct object, otherwise out
        if update.get("object") != "page":
            logger.warning("Message not a object: page, ignoring")
            logger.warning(update)
            self.ignore = True
            return
        print(update)
        msg_info = update["entry"][0]["messaging"][0]
        # Check who sent it
        sender_id = msg_info["sender"]["id"]
        self._chat_id = sender_id
        # Get the msg
        msg = msg_info["message"]
        # get the text and parse it if necessary
        self._parse_command(msg.get("text"))
        # In facebook we have either text or image
        # TODO: in facebook you can pass more than one img at once...
        attachment = msg.get("attachments")
        if attachment is not None:
            at_info = attachment[0]
            # Checked for images and files and seems to work
            url = at_info["payload"]["url"]
            self._file_id = url
            self._text = url.split("?")[0].split("/")[-1]


class FacebookUtil(Backend):
    """This class handles all communications with
    Telegram"""

    _message_class = FacebookMessage
    _max_size = MAX_SIZE

    def __init__(self, PAGE_TOKEN, VERIFY_TOKEN, host="0.0.0.0", port=3000, debug=False):
        if not _HAS_FLASK:
            # Raise the error now
            raise ModuleNotFoundError("No module named 'flask'")

        self.page_access_token = PAGE_TOKEN
        self.verify_token = VERIFY_TOKEN
        self.port = port
        self.host = host
        app = Flask(__name__)
        # Load the listener into the webhook endpoint
        app.add_url_rule("/webhook", "webhook", self.listener, methods=["POST", "GET"])
        self.flask_app = app
        self.debug = debug
        self.action_function = None
        self.auth = {"access_token": self.page_access_token}

    def validate_hook(self):
        """Facebook needs to validate the webhook
        This is a small utility to do so
        """

    def listener(self):
        """Main function flask will use to listen at the webhook endpoint"""
        if request.method == "GET":
            if request.args.get("hub.verify_token") == self.verify_token:
                return request.args.get("hub.challenge")
            else:
                return "incorrect"

        if request.method == "POST":
            msg = self._message_class(request.json)
            self.action_function(msg)
            logger.info(msg)
            # After we have finished return a 200 Ok
            return "All ok"

    def act_on_updates(self, action_function, not_empty=False):
        """Sets the action function to be used by the listener and then
        opens the webhook to wait ofr updates and act on them
        """
        self.action_function = action_function
        self.flask_app.run(host=self.host, port=self.port, debug=self.debug)

    def _get_updates(self, not_empty=False):
        """This class skips get_updates and uses act_on_updates directly"""
        pass

    def send_message(self, text, chat, **kwargs):
        """Sends a message response to facebook
        If the message is greater than MAX_SIZE characters it gets broken into several msgs
        """
        break_char = self._break_msg(text)
        payload = {"message": {"text": text[:break_char]}, "recipient": {"id": chat}}
        response = requests.post(FB_API, params=self.auth, json=payload)
        if len(text) > break_char:
            return self.send_message(text[break_char:], chat, **kwargs)
        return response.json()

    def send_data(self, payload):
        """Sends data to facebook messenger.
        This method uses MultipartEncoder: https://toolbelt.readthedocs.io/
        to stream multipart form-data
        """
        try:
            from requests_toolbelt import MultipartEncoder
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "Install 'requests-toolbelt' to send images and files to facebook"
            ) from e

        encoded_payload = MultipartEncoder(payload)
        header = {"Content-Type": encoded_payload.content_type}
        response = requests.post(FB_API, params=self.auth, data=encoded_payload, headers=header)
        return response.json()

    def send_image(self, img_path, chat):
        """Sends an image to facebook
        Basically the requests form of the curl command here:
        https://developers.facebook.com/docs/messenger-platform/send-messages#url
        """
        img = pathlib.Path(img_path)
        payload = {
            "recipient": json.dumps({"id": chat}),
            "message": json.dumps(
                {"attachment": {"type": "image", "payload": {"is_reusable": True}}}
            ),
            "filedata": (img.stem, img.read_bytes(), f"image/{img.suffix[1:]}"),
        }
        return self.send_data(payload)

    def send_file(self, filepath, chat):
        """Sends a file to fb, similar to send_image"""
        fff = pathlib.Path(filepath)
        payload = {
            "recipient": json.dumps({"id": chat}),
            "message": json.dumps(
                {"attachment": {"type": "file", "payload": {"is_reusable": True}}}
            ),
            "filedata": (fff.name, fff.read_bytes()),
        }
        return self.send_data(payload)


if __name__ == "__main__":
    logger.info("Testing FB Util")
    verify = "your_verify_token"
    app_token = "your_app_key"
    fb_util = FacebookUtil(app_token, verify, debug=True)
    fb_util.act_on_updates(lambda x: print(x))



"""
    Module to read and send content from wikipedia
"""

# TODO use the selected language

import re

import wikipedia

from pybliotecario.components.component_core import Component

# log = logging.getLogger(__name__)
GET_N = re.compile(r"^\d+")
MAX_SIZE = 4000
WIKI_NAME = "WIKIPEDIA"
DEFAULT_LANGUAGE = "EN"


class WikiComponent(Component):
    """
    Read content from wikipedia and send it to instagram
    """

    help_text = """ > Wikipedia module
    /wiki term: search for term in wikipedia, return the summary
    /wiki_full N term: read the full article and return N times the defined summary_size"""

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        wiki_config = self.read_config_section(WIKI_NAME)
        self.summary_size = int(wiki_config.get("msg_size", 1024))
        self.language = wiki_config.get("language", DEFAULT_LANGUAGE)
        wikipedia.set_lang(self.language)

    @classmethod
    def configure_me(cls):
        print(
            f"""
# Wikipedia Module
This is the configuration helper for the wikipedia module
Introduce the length of the msgs you want to obtain (max: {MAX_SIZE}):"""
        )
        summary_size = 0
        while summary_size > MAX_SIZE or summary_size < 1:
            # The max msg in telegram is 4096 UTF char.
            try:
                summary_size = int(input(" > "))
            except ValueError:
                print(f"Please, write a number between 0 and {MAX_SIZE}")
        possible_languages = ["EN", "ES", "IT"]
        language = None
        print("Introduce the default language for Wikipedia pages")
        while language not in possible_languages:
            language = input(
                f" > Possible choices: {possible_languages} (default: {DEFAULT_LANGUAGE}) > "
            )
            if language == "":
                language = DEFAULT_LANGUAGE
        dict_out = {WIKI_NAME: {"msg_size": summary_size, "language": language}}
        return dict_out

    def get_page(self, term):
        """Gets the wikipedia page given by `term`
        If `term` returns a disambiguation page, returns a string with the options
        """
        try:
            wiki_page = wikipedia.page(term)
            return wiki_page
        except wikipedia.exceptions.DisambiguationError:
            other_terms = wikipedia.search(term)
            response = "Found disambiguation page, please try with one of these other terms: "
            response += ", ".join(other_terms)
        except wikipedia.exceptions.PageError:
            response = f"Sorry, no Wikipedia pages were found for this query: {term}"
        self.send_msg(response)
        return None

    def read_summary(self, term):
        """Gets the page given by `term` and returns a summary
        The size of the summary is given by the `summary_size` variable
        """
        wiki_page = self.get_page(term)
        if wiki_page is not None:
            summary = wiki_page.summary[: self.summary_size]
            number_of_msg = int(len(wiki_page.content) / self.summary_size)
            summary += f"\n\n This page will need {number_of_msg} msgs for the whole content"
            self.send_msg(summary)

    def read_msg_fullpage(self, msg):
        """Breaks the msg = "N term to search for" into
        N (number) and the term to search for (string)
        """
        reg_n = GET_N.search(msg)
        if reg_n is None:
            error = f"""For a whole page, write the query in the form /wiki_full N search_term
where N is the number of msgs of size {self.summary_size} you want to receive"""
            self.send_msg(error)
            return 0, None
        n_str = reg_n[0]
        n_int = int(n_str)
        term = msg.replace(n_str, "").strip()
        return n_int, term

    def read_whole_page(self, term, n_msg):
        """
        Given a `term` to search for and a number of msgs to read `n_msg`
        sends the first `n_msg`*`self.summary_size` of the
        corresponding wikipedia page
        """
        wiki_page = self.get_page(term)
        if wiki_page is not None:
            whole_page = wiki_page.content
            total_length = len(whole_page)
            final_length = min(total_length, n_msg * self.summary_size)
            for char in range(0, final_length, self.summary_size):
                response = whole_page[char : char + self.summary_size]
                self.send_msg(response)

    def telegram_message(self, msg):
        """Digest the telegram msg"""
        command = msg.command
        if command == "wiki":
            self.read_summary(msg.text.strip())
        elif command == "wiki_full":
            n_msg, term = self.read_msg_fullpage(msg.text.strip())
            if n_msg > 0:
                self.read_whole_page(term, n_msg)



"""
    Module using the arxiv API for prelogger.info tracking
    update notifications and quick-querying from the Telegram app
"""

from datetime import datetime, timedelta, timezone
import logging
import os

import arxiv

from pybliotecario.components.component_core import Component

logger = logging.getLogger(__name__)


def is_today(time_to_test):
    """
    Checks whether the given time to test corresponds to today
    or to the previous day
    Remember, the previous day for arxiv purposes only starts at 19.00
    """
    # First we need to find out which day is today
    today = datetime.now(timezone.utc)
    wday = today.weekday()  # 0 == Monday
    base_hour = 18

    if wday < 2:
        # If today is monday or tuesday, we should look at thursday/friday for the cutoff
        base_today = today - timedelta(days=4)
    else:  # yesterday
        base_today = today - timedelta(days=2)
    base_today = base_today.replace(hour=base_hour, minute=0, second=0, microsecond=0)

    # Now, if the paper date (pdt) is from before the base time (base_today), that means it is not from today
    return time_to_test >= base_today


def query_recent(category):
    """
    Query the arxiv for the updates of the last day for a given category
    """
    results = arxiv.Search(
        query=category, max_results=75, sort_by=arxiv.SortCriterion.LastUpdatedDate
    ).results()
    elements = []
    for _, element in enumerate(results):
        if is_today(element.published):
            elements.append(element)
    logger.info("Found %d new papers", len(elements))
    return elements


def check_keyword(paper_value, keyword):
    """Check whether the given keywords exist in the given paper value"""
    if isinstance(paper_value, list):
        return any([check_keyword(i, keyword) for i in paper_value])
    str_value = str(paper_value)
    return keyword.casefold().strip() in str_value.casefold().strip()


def filter_results(result_list, filter_dictionary):
    """
    Given a dictionary and a list of results (paper dictionaries)
    returns the results which pass the filter
    The items in the filter_dictionary must be a list
    """  # TODO: save the reason for the query to pass
    lista = []
    for paper in result_list:
        done = False
        for key, keywords in filter_dictionary.items():
            if done:
                break
            try:
                val = getattr(paper, key)
            except AttributeError:
                logger.error("Error trying to read key: %s", key)
                continue

            for keyword in keywords:
                if check_keyword(val, keyword):
                    lista.append(paper)
                    done = True
                    break
    return lista


def url_to_id(arxiv_url):
    """
    Given an arxiv url return the corresponding id
    if arxiv_url is already an id, returns unchanged
    """
    return os.path.basename(arxiv_url)


# Telegram-usable functions
def arxiv_get_pdf(arxiv_id_raw):
    """Downloads a paper from the arxiv given an id"""
    arxiv_id = url_to_id(arxiv_id_raw)
    # First we recover the information about the paper
    try:
        paper = next(arxiv.Search(id_list=[arxiv_id]).results())
    except arxiv.arxiv.HTTPError:
        return None
    return paper.download_pdf("/tmp/")


def arxiv_recent_filtered(categories, filter_dict, abstract=False, max_authors=50):
    """Read all recent papers in each category and save the ones matching the
    requirements of filter_dict"""
    lines = []
    for category in categories:
        tmp = query_recent(category)
        results = filter_results(tmp, filter_dict)
        url = f"https://arxiv.org/list/{category}/new"
        n_new = len(tmp)
        n_int = len(results)
        line = f"{n_new} new papers in [{category}]({url}) "
        if n_int > 0:
            line += f"{n_int} interesting ones:\n"
        else:
            line += "(nothing to be highlighted)."
        for paper in results:
            paper_authors = paper.authors
            if len(paper_authors) > max_authors:
                paper_authors = paper_authors[:max_authors] + ["et al"]
            authors = ", ".join(str(i) for i in paper_authors)
            arxiv_id = paper.get_short_id().replace("v1", "")
            line += f" > {paper.title}: [{arxiv_id}]({paper.links[0]})\n  by {authors}\n"
            if abstract:
                line += paper["summary"]
        lines.append(line)
    return "\n".join(lines)


def arxiv_query_info(arxiv_id_raw):
    """
    Returns extra information about the queried paper
    """
    arxiv_id = url_to_id(arxiv_id_raw)
    paper = next(arxiv.Search(id_list=[arxiv_id]).results())
    authors = [str(i) for i in paper.authors]
    abstract = paper.summary.replace("\n", " ")
    msg = f""" > {arxiv_id}
Title: {paper.title}

Authors: {authors}

Abstract: {abstract}
    """
    return msg


class Arxiv(Component):
    help_text = """ > Arxiv module
    /arxiv arxiv_id: sends information about the given id
    /arxiv_get arxiv_id: sends the PDF for the given id"""

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        arxiv_config = self.read_config_section("ARXIV", telegram_error=False)
        title = arxiv_config.get("title", "")
        abstract = arxiv_config.get("summary", "")
        authors = arxiv_config.get("authors", "")
        self._max_authors = arxiv_config.get("max_authors", 20)
        self.filter_dict = {
            "title": self.split_list(title),
            "summary": self.split_list(abstract),
            "authors": self.split_list(authors),
        }
        self.categories = self.split_list(arxiv_config.get("categories", ""))

    @classmethod
    def configure_me(cls):
        print("")
        print(" # Arxiv module ")
        print("This is the configuration helper for the arxiv module")
        print("First introduce (comma-separated) the categories you are interested in: ")
        categories = input(" > ")
        print(
            "Now introduce (again, comma-separated) the keywords you want to check the titles for"
        )
        title_keywords = input(" > ")
        print("Same for checking the abstract:")
        summary_keywords = input(" > ")
        print("And now authors")
        authors_keywords = input(" > ")
        dict_out = {
            "ARXIV": {
                "categories": categories,
                "title": title_keywords,
                "summary": summary_keywords,
                "authors": authors_keywords,
            }
        }
        return dict_out

    def cmdline_command(self, args):
        """
        Reads the latest arrivals to the arxiv and
        sends a notification if any of the entries fulfill any of
        the requirements set in the config file
        """
        msg = arxiv_recent_filtered(
            self.categories, self.filter_dict, max_authors=self._max_authors
        )
        self.send_msg(msg, markdown=True)
        logger.info("Arxiv information sent")

    def telegram_message(self, msg):
        command = msg.command
        arxiv_id = msg.text.strip()
        if not arxiv_id:
            self.send_msg("This commands needs an argument")
            return
        if command in ("arxivget", "arxiv-get", "arxiv_get"):
            file_send = arxiv_get_pdf(arxiv_id)
            if file_send is None:
                self.send_msg("Error trying to download the paper, please check ID again")
            else:
                self.send_file(file_send, delete=True)
        else:
            msg = arxiv_query_info(arxiv_id)
            self.send_msg(msg)


if __name__ == "__main__":
    import tempfile

    from pybliotecario.pybliotecario import logger_setup

    logger_setup(tempfile.TemporaryFile(), debug=True)
    logger.info("Testing the arxiv component")
    logger.info("Query a category")
    categoria = "hep-ph"
    recent_papers = query_recent(categoria)
    dict_search = {
        "title": ["Higgs", "neutrinos"],
        "authors": ["Cruz-Martinez"],
        "summary": ["VBF"],
    }
    filter_papers = filter_results(recent_papers, dict_search)
    logger.warning(filter_papers)
    tlg_msg = arxiv_recent_filtered([categoria], dict_search)
    logger.info(tlg_msg)

#     logger.info("Test download")
#     test_id = "1802.02445"
#     name = arxiv_get_pdf(test_id)
#     os.remove(name)



"""
    Tests the stocks module
"""
import json
from pybliotecario.pybliotecario import main
from pybliotecario.backend import TestUtil
from .conftest import generate_fake_config


def test_cmdline_usage(tmpfile):
    """Generates a fake json for stock watching
    and then sends it to the pybliotecario
    """
    # Prepare the json file
    json_example = {"AAPL": {"below": 5, "above": 5}}
    tmp = tmpfile.with_suffix(".json")
    with open(tmp, "w") as f:
        json.dump(json_example, f)
    args = [f"--stock_watcher", str(tmp)]
    # Create the fake backend
    test_util = TestUtil(communication_file=tmpfile)
    fake_config = generate_fake_config(tmpfile)
    main(cmdline_arg=args, tele_api=test_util, config=fake_config)
    messages = tmpfile.read_text().strip().split("\n")
    # We should get just one message saying that AAPL is above the threshold
    assert len(messages) == 1
    assert "price is above" in messages[0]



#!/usr/bin/env python3
"""
    Main script for the pybliotecario program
    for command line invocation
"""
import logging
from pathlib import Path
import sys

# Modify argument_parser.py to read new arguments
from pybliotecario.argument_parser import parse_args
from pybliotecario.backend import FacebookUtil, TelegramUtil, TestUtil
from pybliotecario.core_loop import main_loop
from pybliotecario.customconf import CustomConfigParser, default_config_path
import pybliotecario.on_cmdline as on_cmdline

logger = logging.getLogger()


def read_config(config_file=None):
    """Reads the pybliotecario config file and uploads the global configuration
    By default looks always in the default file path (in XDG_CONFIG_HOME) and the current folder
    """
    default_file_path = default_config_path()
    # Checks as well (with lower priority) for ~/.pybliotecario.ini for backwards compatibility
    old_path = Path.home() / ".pybliotecario.ini"
    config_files = [old_path, default_file_path, default_file_path.name]
    if old_path.exists():
        logger.error(
            f"Deprecation notice: ~/.pybliotecario.ini is now deprecated, please move the configuration to {default_file_path}"
        )

    if config_file is not None:
        config_files.append(config_file)
    config = CustomConfigParser()
    config.read(config_files, encoding="UTF-8")
    # Add a custom paster to this config
    if config and config.defaults():
        return config
    print("Before using this program you need to run the --init option in order to configure it")
    sys.exit(-1)


def logger_setup(filename, debug=False):
    """
    Send all logger output to file
    But make sure that errors and warning also go
    to stdout
    """
    # Set the formatter
    formatter = logging.Formatter("[%(levelname)s:%(asctime)s] - %(message)s", "%d/%m/%y %H:%M")
    if debug:
        file_handler = logging.StreamHandler(sys.stdout)
    else:
        # Error handler (not necessary when debug is on):
        error_handler = logging.StreamHandler(sys.stdout)
        error_handler.setFormatter(formatter)
        error_handler.setLevel(logging.WARNING)
        logger.addHandler(error_handler)
        # Default handler
        file_handler = logging.FileHandler(filename, "a", encoding="UTF-8")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    logger.setLevel(logging.INFO)


def main(cmdline_arg=None, tele_api=None, config=None):
    """Driver of the pybliotecario

    For debugging purposes,
    If the arguments or the backend is inputted to this function
    it will override any command line arguments.

    >>> from pybliotecario.pybliotecario import main
    >>> from pybliotecario.backend import TestUtil
    >>> args = ["This is only a test"]
    >>> tt = TestUtil("/tmp/test.txt")
    >>> main(cmdline_arg=args, tele_api=tt)
    """

    # Parse the input arguments and the configuration file
    if cmdline_arg is None:
        cmdline_arg = sys.argv[1:]
    args = parse_args(cmdline_arg)

    # Parse the configuration file
    if config is None:
        config = read_config(args.config_file)
        defaults = config.defaults()
        main_folder = defaults.get("main_folder")
        if not main_folder:
            logger.warning(
                "No 'default:main_folder' option set in %s, using /tmp/", args.config_file
            )
            main_folder = "/tmp/"
        logger_setup(main_folder + "/info.log", debug=args.debug)

    logger.info("Initializing the pybliotecario")

    # Check the backend the pybliotecario should be using
    if tele_api is None:
        if args.backend.lower() == "telegram":
            api_token = config.defaults().get("token")
            if not api_token:
                logger.error(
                    "No 'default:token' option set in %s, run --init option", args.config_file
                )
                sys.exit(-1)

            tele_api = TelegramUtil(api_token, debug=args.debug)
        elif args.backend.lower() == "test":
            tele_api = TestUtil("/tmp/test_file.txt")
        elif args.backend.lower() == "facebook":
            try:
                fb_config = config["FACEBOOK"]
            except KeyError:
                raise ValueError("No facebook section found for facebook in pybliotecario.ini")
            verify_token = fb_config.get("verify")
            app_token = fb_config.get("app_token")
            tele_api = FacebookUtil(app_token, verify_token, debug=args.debug)
            # Check whether we have chat id
            chat_id = fb_config.get("chat_id")
            if chat_id is not None:
                config.set("DEFAULT", "chat_id", chat_id)

    try:
        on_cmdline.run_command(args, tele_api, config)
    except ModuleNotFoundError as e:
        logger.error("In order to use this option you need to install the module '%s'", e.name)

    if args.daemon:
        logger.info("Activating main loop")
        while True:
            main_loop(tele_api, config=config, clear=args.clear_incoming)
            if args.exit_on_msg:
                break


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
    Telegram backend
"""

import json
import logging
from time import sleep
import urllib

import requests

from pybliotecario.backend.basic_backend import Backend, Message

TELEGRAM_URL = "https://api.telegram.org/"

logger = logging.getLogger(__name__)

# Keys included in telegram chats that basically are telling you to ignore it
IGNOREKEYS = {"new_chat_participant", "left_chat_participant", "sticker", "game", "contact"}


def log_request(status_code, reason, content):
    """Log the status of the send requests"""
    result = f"Request sent, status code: {status_code} - {reason}: {content}"
    logger.info(result)


class TelegramMessage(Message):
    """Telegram implementation of the Message class"""

    _type = "Telegram"
    _group_info = None

    def _parse_update(self, update):
        """Receives an update in the form of a dictionary (that came from a json)
        and fills the message attributes
        """
        keys = update.keys()
        # First check whether this is a message, edited message or a channel post
        msg_types = ["message", "edited_message", "edited_channel_post"]
        msg = None
        for msg_type in msg_types:
            if msg_type in keys:
                msg = msg_type
        if msg is None:
            logger.warning(f"Message not in {msg_types}, ignoring")
            logger.warning(update)
            self.ignore = True
            return
        message = update[msg]
        # Get the keys of the message (and check whether it should be ignored)
        msg_keys = message.keys()
        if set(msg_keys) & IGNOREKEYS:
            self.ignore = True
            return
        # Now get the chat data and id
        chat_data = message["chat"]
        self._chat_id = chat_data["id"]
        # I believe from and chat_data are only different when using a group
        # but the pybliotecario has not really been tested (or used) in groups...
        from_data = message.get("from", chat_data)

        # Populate the user (in the list, last has more priority)
        username = "unknown_user"
        for user_naming in ["last_name", "first_name", "username"]:
            username = from_data.get(user_naming, username)
        self._username = username

        # Check the filetype
        text = None
        if "photo" in message:
            # If it is a photo, get the file id and use the caption as the title
            photo_data = message["photo"][-1]
            self._file_id = photo_data["file_id"]
            text = message.get("caption", "untitled")
        elif "document" in message:
            # If it is a document, telegram gives you everything you need
            file_dict = message["document"]
            self._file_id = file_dict["file_id"]
            text = message.get("caption", None)
            if text is None:
                # Use the file name
                text = file_dict["file_name"]
        else:
            # Normal text message
            text = message.get("text", "")

        # In Telegram we can also have groups
        if "group" in chat_data:
            self._group_info = chat_data

        # Finally, with the piece of text left, parse the possible command
        # _parse_command will fill both text and command
        self._parse_command(text)

    @property
    def is_group(self):
        """Returns true if the message was from a group"""
        return self._group_info is not None


class TelegramUtil(Backend):
    """This class handles all communications with
    Telegram"""

    _message_class = TelegramMessage

    def __init__(self, TOKEN, debug=False, timeout=300):
        self.offset = None
        self.debug = debug
        self.timeout = timeout
        # Build app the API urls
        base_URL = TELEGRAM_URL + f"bot{TOKEN}/"
        self.base_fileURL = TELEGRAM_URL + f"file/bot{TOKEN}/"
        self.send_msg = base_URL + "sendMessage"
        self.send_img = base_URL + "sendPhoto"
        self.send_doc = base_URL + "sendDocument"
        self.get_msg = base_URL + "getUpdates"
        self.get_file = base_URL + "getFile"

    def __make_request(self, url):
        """Returns the response for a given url
        In case of timeout, emulate an empty response
        """
        content = '{ "ok": true, "result": [] }'
        try:
            response = requests.get(url, timeout=self.timeout)
            content = response.content.decode("utf-8")
        except requests.exceptions.Timeout:
            logger.warning("Timeout. Sleeping for 2 minutes...")
            # After a timeout, sleep for 1 minute
            sleep(60)
        except requests.exceptions.ConnectionError as e:
            logger.error(e)
        return content

    def __get_json_from_url(self, url):
        """Return the json response of a given url"""
        content_response = self.__make_request(url)
        res_json = json.loads(content_response)
        return res_json

    def __re_offset(self, updates):
        """Updates the offset
        i.e., resets don't ask Telegram server for old messages
        """
        if not updates:  # len(updates) == 0
            return
        li = []
        for update in updates:
            li.append(int(update["update_id"]))
        self.offset = max(li) + 1

    def _get_filepath(self, file_id):
        """Given a file id, retrieve the URI of the file
        in the remote server
        """
        url = self.get_file + f"?file_id={file_id}"
        jsonret = self.__get_json_from_url(url)
        # was it ok?
        if jsonret["ok"]:
            fpath = jsonret["result"]["file_path"]
            return self.base_fileURL + fpath
        else:
            logger.info(jsonret["error_code"])
            logger.info("Here's all the information we have on this request")
            logger.info("This is the url we have used")
            logger.info(url)
            return None

    def _get_updates(self, not_empty=False):
        """
        Returns a json with the last messages the bot has received
        when an offset is found, previous msg are not retrieved
        We use longpolling to keep the connection open for timeout seconds

        If not_empty = True, this function will only return when a message arrives
        """
        url = f"{self.get_msg}?timeout={self.timeout}"
        if self.offset:
            url += f"&offset={self.offset}"
        updates = self.__get_json_from_url(url)

        if not updates and not_empty:
            return self._get_updates(not_empty=True)

        if self.debug:
            logger.info("Request url: %s", url)
            logger.info("Obtained updates: %s", updates)

        try:
            result = updates["result"]
        except Exception as e:
            # in case of ANY exception, just logger.info it out and let the program run
            logger.error("Error: ")
            logger.error(str(e))
            logger.error("List of updates: ")
            logger.error(updates)
            return []

        self.__re_offset(result)
        return result

    def send_message(self, text, chat, markdown=False, **kwargs):
        """Send a message to a given chat"""
        text = urllib.parse.quote_plus(text)
        url = f"{self.send_msg}?text={text}&chat_id={chat}"
        if markdown:
            url += "&parse_mode=markdown"
        content = self.__make_request(url)
        # Sometimes, when using markdown, the update might fail
        if markdown and not json.loads(content).get("ok"):
            # If that's the case, try to resend without md
            self.send_message(text, chat, **kwargs)

    def send_image(self, img_path, chat):
        """Send an image to a given chat"""
        data = {"chat_id": chat}
        with open(img_path, "rb") as img:
            files = {"photo": ("picture.jpg", img)}  # Here, the "rb" thing
            blabla = requests.post(self.send_img, data=data, files=files)
        log_request(blabla.status_code, blabla.reason, blabla.content)

    def send_file(self, filepath, chat):
        data = {"chat_id": chat}
        files = {"document": (filepath.name, filepath.open("rb"))}
        blabla = requests.post(self.send_doc, data=data, files=files)
        log_request(blabla.status_code, blabla.reason, blabla.content)

    def send_file_by_url(self, file_url, chat):
        """
        Sends a file using the telegram api which allows to send by url
        it only works for images/pdf, and not all pdfs
        """
        data = {"chat_id": chat, "document": file_url}
        blabla = requests.post(self.send_doc, data=data)
        logger.info(blabla.status_code, blabla.reason, blabla.content)

    def download_file(self, file_id, file_path):
        """Download file defined by file_id to given file_path"""
        file_url = self._get_filepath(file_id)
        if not file_url:
            return None
        n = 0
        # If the file already exists, iterate it by adding a nX
        while file_path.exists():
            new_name = f"n{n}-{file_path.name}"
            file_path = file_path.parent / new_name
            n += 1
        return urllib.request.urlretrieve(file_url, file_path)


if __name__ == "__main__":
    logger.info("Testing TelegramUtil")
    token = "must put a token here to test"
    ut = TelegramUtil(token, debug=True)
    # noinspection PyProtectedMember
    results = ut._get_updates()
    for res in results:
        logger.info("Complete json:")
        logger.info(res)
        msg_example = res["message"]
        chat_id = msg_example["chat"]["id"]
        txt = msg_example["text"]
        logger.info("Message from %s: %s", chat_id, txt)
        ut.send_message("Message received", chat_id)
    ut.timeout = 1
    # noinspection PyProtectedMember
    ut._get_updates()  # Use the offset to confirm updates



"""
    This module manages the core loop of the pybliotecario
    when it is called with daemon mode -d
"""
from datetime import datetime
import logging
from pathlib import Path

import pybliotecario.on_cmd_message as on_cmd_message

logger = logging.getLogger(__name__)

# After which number of continuous exceptions do we actually fail
_FAILTHRESHOLD = 20


def _monthly_folder(base_main_folder):
    """Receives a path object with the base main folder
    and returns the monthly folder (also a Path object)"""
    main_folder = base_main_folder / "data"
    ahora = datetime.now()
    y = str(ahora.year)
    m = ahora.strftime("%B")
    folder_name = main_folder / y / m
    folder_name.mkdir(exist_ok=True, parents=True)
    return folder_name


def _write_to_daily_log(main_folder, msg):
    """Write the msg to the daily log"""
    folder = _monthly_folder(main_folder)
    day = datetime.now().day
    file_name = folder / f"{day}.log"
    with file_name.open("a+", encoding="utf-8") as f:
        f.write(msg)
        f.write("\n")


def still_alive():
    from random import randint

    sentences = [
        "Pong",
        "This was a triumph",
        "HUGE SUCCESS",
        "It's hard to overstate my satisfaction",
        "There's no sense crying over every mistake",
        "You just keep on trying till you run out of cake",
        "I'm being so sincere right now",
        "You tore me to pieces",
        "These points of data make a beautiful line",
        "We're out of beta, we're releasing on time",
        "Think of all the things we learned",
        "Go ahead and leave me",
        "I think I prefer to stay inside",
        "This cake is great, it's so delicious and moist",
        "When I look out there it makes me glad I'm not you",
        "Believe me, I am still alive",
        "I'm doing science and I'm still alive",
        "I feel FANTASTIC and I'm still alive",
        "While you're dying I'll be still alive",
        "When you're dead I will be still alive",
    ]
    r = randint(0, len(sentences) - 1)
    return sentences[r]


def main_loop(tele_api, config=None, clear=False):
    """
    This function defines a "listener" which will wait for messages
    in the form of pybliotecario.Message objects and will act on them

    The task of the tele_api is then to call the generated function with an
    instance of the message as the argument.
    This allows the tele_api to do things asynchronously if needed be
    """
    main_folder = Path(config["DEFAULT"]["main_folder"])
    accepted_ids = config.getidlist("DEFAULT", "chat_id")
    main_id = config.getmainid("DEFAULT", "chat_id")
    chivato = config.getboolean("DEFAULT", "chivato", fallback=False)

    global except_counter
    except_counter = 0

    # Generate the function to act on Messages
    def act_on_message(message):
        """This function receives a pybliotecario.Message and
        an actor object and calls act_on_telegram_command as required"""
        global except_counter
        # Check whether the message should be ignored
        try:  # Wrap everything on a try-except block which will not crash if clear=True
            if message.ignore:
                except_counter = 0
                return
            chat_id = message.chat_id

            # In "chivato" mode, send a message to the main chat_id if sender is not recognized
            if chivato and chat_id not in accepted_ids:
                chivato_msg = f"""El usuario @{message.username} ({chat_id=}) ha enviado el siguiente mensaje:
{message.raw}"""
                tele_api.send_message(chivato_msg, main_id)

            if message.is_command:
                # Call the selected command and act on the message
                on_cmd_message.act_on_telegram_command(tele_api, message, config)
            elif message.is_file:
                # If the message is a file, save the file and we are done
                file_name = message.text.replace(" ", "")
                file_path = _monthly_folder(main_folder) / file_name
                result = tele_api.download_file(message.file_id, file_path)
                if result:
                    tele_api.send_message("¡Archivo recibido y guardado!", chat_id)
                    logger.info("File saved to %s", file_path)
                else:
                    tele_api.send_message("There was some problem with this, sorry", chat_id)
                    logger.info(message)
                    logger.warning("There was a problem with this update")
            else:
                # Otherwise just save the msg to the log and send a funny reply
                _write_to_daily_log(main_folder, message.text)
                random_msg = still_alive()
                tele_api.send_message(random_msg, chat_id)
            except_counter = 0
        except Exception as e:
            logger.error(f"This message produced an exception: {e}")
            if clear and except_counter < _FAILTHRESHOLD:
                except_counter += 1
                # Ignore exceptions until we reach the threshold
                logger.info(message)
                logger.info("Going for the next message")
            else:
                raise e

    tele_api.act_on_updates(act_on_message, not_empty=True)



"""
    Tests the test backend
"""

from pybliotecario.backend import TestUtil

_FAKEMSGS = ["This is only a test", "Hola, caracola"]


def test_get_updates_fakemsg():
    """Checks that get updates returns messages
    with the input message inside
    """
    test_util = TestUtil(fake_msgs=_FAKEMSGS)
    updates = test_util._get_updates()
    # Check that indeed the update contains only the given message
    assert len(updates) == len(_FAKEMSGS)
    for update, msg in zip(updates, _FAKEMSGS):
        assert update["message"]["text"] == msg


def test_get_updates_tmpfile(tmpfile):
    """Check that if we write updates to the temporary file we get them back"""
    test_util = TestUtil(communication_file=tmpfile)
    tmpfile.write_text("\n".join(_FAKEMSGS))
    updates = test_util._get_updates()
    assert len(updates) == len(_FAKEMSGS)
    for update, msg in zip(updates, _FAKEMSGS):
        assert update["message"]["text"] == msg


def test_act_on_update():
    """ Check that act on update works as expected """
    test_util = TestUtil(fake_msgs=["/test command"])

    # Action function
    def action(msg):
        assert msg.text == "command"
        assert msg.is_command
        assert msg.command == "test"

    test_util.act_on_updates(action)


def test_is_msg_in_file(tmpfile):
    """Check that the utlity to check for a msg in the
    communication file is working"""
    msg = "This is only a test"
    test_util = TestUtil(tmpfile)
    tmpfile.write_text(msg)
    assert test_util.is_msg_in_file(msg)
    assert not test_util.is_msg_in_file("Something different")


def test_send_msg():
    """Checks that upon sending a message it gets returned"""
    test_util = TestUtil()
    msg = "Hola, caracola"
    # test util does not take a chat id for now
    ret = test_util.send_message(msg, None)
    assert ret == msg


def test_send_msg_tmpfile(tmpfile):
    """Checks that upon sending a message it gets written to the
    given file"""
    test_util = TestUtil(tmpfile)
    msg = "Hola, caracola"
    # test util does not take a chat id for now
    test_util.send_message(msg, None)
    assert test_util.is_msg_in_file(msg)



"""
    Base/abstract backend classes:

    - Message
    - Backend

    Each backend should implement its own message type and inherit from backend
"""

from abc import ABC, abstractmethod
import json
import logging
import urllib

logger = logging.getLogger(__name__)


class Message(ABC):
    """
    Base message class

    Any implementation of the message should read the incoming update
    and save the following information in the appropiate attributes:
     - chat id
     - username of the sender
     - text of the msg
     - command (if any)
     - file id (if any)
     - ignore (if the msg is to be ignored)
    through the ``_parse_update`` method.
    """

    _type = "Abstract"
    _original = None

    def __init__(self, update):
        self._chat_id = None
        self._username = None
        self._command = None
        self._file_id = None
        self._text = None
        self._ignore = False
        self._raw = None
        self._original = update
        self._parse_update(update)
        # After the information is parsed, log the message!
        logger.info("New message: %s", self)

    def __str__(self):
        rep = ["chat_id", "username", "command", "file_id", "text", "ignore"]
        msg_dict = {}
        for r in rep:
            msg_dict[r] = getattr(self, r)
        return json.dumps(msg_dict)

    def _parse_command(self, text):
        """Separate messages containing commands into the command and text attributes
        a command is a message which starts with /"""
        if not text.startswith("/"):
            self._command = None
            self._text = text
            return
        separate_command = text.split(" ", 1)
        # Remove the / from the command
        command = separate_command[0][1:]
        # Absorb the @ in case it is a directed command!
        if "@" in command:
            command = command.split("@")[0]
        # Check whether the command comes alone or has arguments
        if len(separate_command) == 1:
            text = ""
        else:
            text = separate_command[1]
        self._command = command
        self._text = text

    @abstractmethod
    def _parse_update(self, update):
        """Parse the update and fill in all entries"""
        return None

    @property
    def raw(self):
        """Return the raw message, understood as command + text"""
        original = ""
        if self.is_command:
            original += f"/{self.command}"
        if self.text is not None:
            original += self.text
        if self.is_file:
            original += " (contains file)"
        return original

    @property
    def chat_id(self):
        """Returns the chat id"""
        return self._chat_id

    @property
    def username(self):
        """Returns the username"""
        return self._username

    @property
    def text(self):
        """Returns the content of the message"""
        return self._text

    @property
    def is_command(self):
        """Returns true if the message is a command"""
        return self.command is not None

    @property
    def command(self):
        """Returns the command contained in the message"""
        return self._command

    @property
    def is_file(self):
        """Returns true if the message is a file"""
        return self.file_id is not None

    @property
    def file_id(self):
        """Returns the id of the file"""
        return self._file_id

    @property
    def has_arguments(self):
        """Returns true if the message is a command with arguments"""
        return self.is_command and self.text is not None

    @property
    def ignore(self):
        """Returns true if the message is to be ignored"""
        return self._ignore

    @ignore.setter
    def ignore(self, val):
        self._ignore = val


class Backend(ABC):
    """
    Main backend class for inheriting.

    It provides a number of base functions and wrappers

    The minimum set of methods and properties the backend must define are:

    _message_class: a reference to the Message class of the backend
    _get_updates: a method that should return a list of updates to act upon
    send_message: a method to communicate messages

    Others:
        - send_image
        - send_file
        - download_file

    """

    _max_size = 99999

    @abstractmethod
    def _get_updates(self, not_empty=False):
        """Retrieve updates as a list"""
        return []

    def raw_updates(self):
        """Returns a raw version of the updates as implemented by the child class"""
        return self._get_updates(not_empty=True)

    @abstractmethod
    def send_message(self, text, chat, **kwargs):
        """Sends a message to the chat"""

    @property
    @abstractmethod
    def _message_class(self):
        pass

    def _break_msg(self, msg):
        """Find the position of the closest whitespace to size in order to break a given msg
        It prioritizes newlines, then spaces, then just returns size
        (but only from half the msg onwards)
        """
        if len(msg) < self._max_size:
            return self._max_size
        available_msg = msg[: self._max_size]
        for break_char in ["\n", " "]:
            last_break = available_msg.rindex(break_char)
            if last_break > int(self._max_size / 2):
                return last_break
        return self._max_size

    def act_on_updates(self, action_function, not_empty=False):
        """
        Receive the input using _get_updates, parse it with
        the message class and act in consequence
        """
        all_updates = self._get_updates(not_empty=not_empty)
        for update in all_updates:
            msg = self._message_class(update)
            action_function(msg)

    def send_image(self, img_path, chat):
        """Sends an image"""
        logger.error("This backend does not implement sending images")

    def send_file(self, filepath, chat):
        """Sends a file"""
        logger.error("This backend does not implement sending files")

    def download_file(self, file_id, file_path):
        """Downloads a file using urllib.
        Understands file_id as the url
        """
        return urllib.request.urlretrieve(file_id, file_path)



"""
    Module to deal with reaction pics
"""

# TODO:
#   Make sure you don't save img1.png img1.jpg

import glob
import logging
import os
import pathlib

from pybliotecario.components.component_core import Component

log = logging.getLogger(__name__)
REACTIONS = "reactions"


def list_content_folder_as_str(folder):
    """List the stem of all the files of a folder as a str"""
    reaction_wild = "{0}/*".format(folder)
    reaction_content = glob.glob(reaction_wild)
    files_found = [pathlib.Path(i).stem for i in reaction_content]
    files_str = ", ".join(files_found)
    return files_str


def look_for_file(folder, filename):
    """Receives a `filename` without the extension and looks
    whether it exists in `folder` (with any extension)"""
    reaction_wild = "{0}/{1}.*".format(folder, filename)
    reaction_content = glob.glob(reaction_wild)
    return reaction_content


class Reactions(Component):
    """
        The idea is to upload reactions with:
        /reaction-save blabla (and an image given)
    which will be stored in the folder .pybliotecario/reactions/blabla.png
    and then two options:
        /reaction blabla (which would return the blabla.png image
    and /reaction-list (which would list all files in the reactions' folder)
    """

    help_text = """ > Reactions module
    /reaction_save reaction_name: save an image with name reaction_name
    /reaction_list: list all reactions
    /reaction reaction_name: sends the reaction given by reaction_name """

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        self.reaction_folder = "{0}/{1}".format(self.main_folder, REACTIONS)
        os.makedirs(self.reaction_folder, exist_ok=True)

    def list_reactions(self):
        """List the reaction pictures saved in the computer"""
        files_str = list_content_folder_as_str(self.reaction_folder)
        out_msg = "Reaction pics: {0}".format(files_str)
        self.send_msg(out_msg)

    def save_reactions(self, msg):
        """Saves the reaction within msg to the reaction folder"""
        file_name = msg.text.replace(" ", "")
        file_path = "{0}/{1}".format(self.reaction_folder, file_name)
        file_id = msg.file_id
        self.telegram.download_file(file_id, file_path)
        self.send_msg("Reaction image {0} correctly saved".format(file_name))

    def send_reaction(self, name):
        """Check whether the file `name` is in the reaction folder and,
        if it is, send it back. it can have any extension!"""
        files = look_for_file(self.reaction_folder, name)
        if not files:
            self.send_msg("Error: reaction '{0}' not found".format(name))
            return
        for reaction in files:
            self.send_img(reaction)

    def telegram_message(self, msg):
        command = msg.command
        if command == "reaction":
            self.send_reaction(msg.text.strip())
        elif command == "reaction_list":
            self.list_reactions()
        elif command == "reaction_save":
            self.save_reactions(msg)



"""
    This module uses the OpenWeather API to provide
    weather information for given location/given times
    using: https://github.com/csparpa/pyowm
"""

from datetime import datetime
import logging

from pyowm import OWM

from pybliotecario.components.component_core import Component

log = logging.getLogger(__name__)


class MyOpenWeather:
    """Wrapper to the pyowm 3 API"""

    def __init__(self, weather_api, **kwargs):
        owm = OWM(weather_api, **kwargs)
        self._owm = owm
        self._mgr = owm.weather_manager()

    def check_current_weather(self, location):
        """Checks the current weather at a specific location"""
        weather = self._mgr.weather_at_place(location).weather

        temp = weather.temperature(unit="celsius")["temp"]
        feels_like = weather.temperature(unit="celsius")["feels_like"]
        stat = weather.detailed_status.lower()
        humid = weather.humidity
        city = location.split(",")[0]

        return f"""Right now we have {stat} in {city}
the current temperature is {temp}°C (humidity={humid}%) and it feels like {feels_like}"""

    def _get_forecast(self, location):
        """Gets the full forecast for a given location"""
        forecast = self._mgr.forecast_at_place(location, "3h")
        return forecast.forecast

    # Telegram-usable wrapper
    def will_it_rain(self, location, hour_range=None):
        """
        Checks the forecast for rain today at the given time range
        """
        if hour_range is None:
            hour_range = (0, 23)
        now = datetime.now()
        min_t = now.replace(hour=int(hour_range[0]))
        max_t = now.replace(hour=int(hour_range[1]))
        forecast = self._get_forecast(location)
        # Loop over the weather we have
        rain_h = []
        rain_p = []
        for weather in forecast:
            if weather.precipitation_probability < 0.4:
                continue
            weather_time = datetime.fromtimestamp(weather.ref_time)
            if weather_time < min_t or weather_time >= max_t:
                continue
            rain_h.append(weather_time.hour)
            rain_p.append(int(weather.precipitation_probability * 100))

        if rain_h:
            rain_str = [f"{i}h ({j}%)" for i, j in zip(rain_h, rain_p)]
            rain_times = ", ".join(rain_str)
            msg = "I find possible rain at the following times: {0} ".format(rain_times)
        else:
            msg = "I can see no rain in my foreseeable future"
        return msg


class Weather(Component):
    """
    Checks the weather forecast at a given location
    at some given times as well as the current forecast at the
    given location and reports to Telegram the findings
    """

    def __init__(self, telegram_object, configuration=None, **kwargs):
        super().__init__(telegram_object, configuration=configuration, **kwargs)
        weather_config = self.read_config_section("WEATHER")
        self.api_key = weather_config.get("api")
        self.weather_location = weather_config.get("location")
        times_str = weather_config.get("times")
        self.check_times = self.split_list(times_str)

    @classmethod
    def configure_me(cls):
        print("")
        print(" # WEATHER MODULE # ")
        print("In order to configure the weather module, you need an API token from OpenWeatherMap")
        print("Go here in order to get one: https://openweathermap.org/appid")
        print(
            "Introduce below the API token, leave empty to skip the configuration of the weather module"
        )
        api = input(" > ")
        if api.strip() == "":
            return None
        print("At which times (comma-separated) do you want to check the weather?")
        times = input(" > ")
        print("In which city?")
        location = input(" > ")
        dict_out = {"WEATHER": {"api": api, "location": location, "times": times}}
        return dict_out

    def cmdline_command(self, args):
        # instantiate open weather object
        open_weather = MyOpenWeather(self.api_key)
        # ask whether it will rain in the future
        msg_rain = open_weather.will_it_rain(self.weather_location, self.check_times)
        # ask what is the current weather
        msg_current = open_weather.check_current_weather(self.weather_location)
        # join everything on a nice msg and send
        weather_msg = "{1}\n{0}".format(msg_rain, msg_current)
        self.send_msg(weather_msg)
        log.info("Weather information sent")


if __name__ == "__main__":
    import tempfile

    from pybliotecario.pybliotecario import logger_setup

    logger_setup(tempfile.TemporaryFile(), debug=True)
    log.info("Testing weather")
    weather_api = "<weather API>"
    city = "Milano, IT"
    owm = MyOpenWeather(weather_api)
    crw = owm.check_current_weather(city)
    log.info(crw)
    msg = owm.will_it_rain(city)
    log.info(msg)
